import 'dart:async';
import 'dart:io';
import 'package:code_id_network/code_id_network.dart';
import 'package:connectivity_plus/connectivity_plus.dart';
import 'package:dio/io.dart';

import 'package:flutter/material.dart';
import 'package:fpdart/fpdart.dart';

typedef RetryEvaluator = FutureOr<bool> Function(DioError error, int attempt);

class NetworkService implements INetworkService {
  late final Dio _dio;
  late final Connectivity connectivity;
  String? method;
  Duration? connectTimeout;
  Duration? receiveTimeout;
  Duration? sendTimeout;
  String baseUrl;
  Map<String, dynamic>? queryParameters;
  Map<String, dynamic>? extra;
  Map<String, dynamic>? headers;
  ResponseType? responseType = ResponseType.json;
  String? contentType;
  bool Function(int?)? validateStatus;
  bool? receiveDataWhenStatusError;
  bool? followRedirects;
  int? maxRedirects;
  List<int> Function(String, RequestOptions)? requestEncoder;
  String Function(List<int>, RequestOptions, ResponseBody)? responseDecoder;
  ListFormat? listFormat;
  bool? persistentConnection;
  NetworkService({
    this.baseUrl = '',
    this.method,
    this.connectTimeout,
    this.receiveTimeout,
    this.sendTimeout,
    this.queryParameters,
    this.extra,
    this.headers,
    this.responseType = ResponseType.json,
    this.contentType,
    this.validateStatus,
    this.receiveDataWhenStatusError,
    this.followRedirects,
    this.maxRedirects,
    this.requestEncoder,
    this.responseDecoder,
    this.listFormat,
    this.persistentConnection,
    Dio? dio,
  }) : _dio = dio ?? Dio() {
    final BaseOptions options = BaseOptions(
      connectTimeout: connectTimeout,
      receiveTimeout: receiveTimeout!,
      sendTimeout: sendTimeout!,
      headers: headers,
      baseUrl: baseUrl,
      contentType: contentType,
      extra: extra,
      followRedirects: followRedirects,
      listFormat: listFormat,
      maxRedirects: maxRedirects,
      method: method,
      queryParameters: queryParameters,
      receiveDataWhenStatusError: receiveDataWhenStatusError,
      requestEncoder: requestEncoder,
      responseDecoder: responseDecoder,
      responseType: responseType,
      persistentConnection: persistentConnection,
      // setRequestContentTypeWhenNoPayload: setRequestContentTypeWhenNoPayload,
      validateStatus: validateStatus,
    );
    _dio.httpClientAdapter = IOHttpClientAdapter()
      ..onHttpClientCreate = (HttpClient client) {
        client.badCertificateCallback =
            (X509Certificate cert, String host, int port) {
          return baseUrl.contains(host);
        };
        return client;
      };

    _dio.options = options;

    connectivity = Connectivity();
  }

  @override
  void addInterceptors(Iterable<Interceptor> interceptors) {
    _dio.interceptors.addAll(interceptors);
  }

  @override
  Future<Either<NetworkError, dynamic>> getHttp({
    required String path,
    int retry = 0,
    String? parameter,
    Map<String, dynamic>? queryParameter,
    Map<String, dynamic>? header,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        final Map<String, dynamic> dioHeader = _dio.options.headers;
        final Map<String, dynamic> headers = {
          'content-type': ContentType.json.mimeType,
          'Accept': ContentType.json.mimeType,
        };
        if (header != null) {
          headers.addAll(header);
        }
        dioHeader.addAll(headers);
        _dio.options.headers = dioHeader;

        final Response response = await _requestGet(
          retry: retry,
          path: '$path${parameter ?? ""}',
          queryParameter: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());
        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  @override
  Future<Either<NetworkError, dynamic>> postHttp({
    required String path,
    int retry = 0,
    String? parameter,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    String? contentType,
    Map<String, dynamic>? header,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        if (content is! FormData) {
          final Map<String, dynamic> dioHeader = _dio.options.headers;
          final Map<String, dynamic> headers = {
            'content-type': contentType ?? ContentType.json.mimeType,
            'Accept': contentType ?? ContentType.json.mimeType,
          };
          if (header != null) {
            headers.addAll(header);
          }
          dioHeader.addAll(headers);
          _dio.options.headers = dioHeader;
        }
        final Response response = await _requestPost(
          retry: retry,
          path: '$path${parameter ?? ""}',
          queryParameter: queryParameter,
          content: content,
          onSendProgress: onSendProgress,
          onReceiveProgress: onReceiveProgress,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());
        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  @override
  Future<Either<NetworkError, dynamic>> putHttp({
    required String path,
    int retry = 0,
    String? parameter,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    String? contentType,
    Map<String, dynamic>? header,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        // await baseStorage.openBox(StorageConstants.base);
        final Map<String, dynamic> dioHeader = _dio.options.headers;
        final Map<String, dynamic> headers = {
          'content-type': ContentType.json.mimeType,
          'Accept': ContentType.json.mimeType,
        };
        if (header != null) {
          headers.addAll(header);
        }
        dioHeader.addAll(headers);
        _dio.options.headers = dioHeader;
        final Response response = await _requestPut(
          retry: retry,
          path: path,
          content: content,
          onSendProgress: onSendProgress,
          queryParameter: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());

        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  @override
  Future<Either<NetworkError, dynamic>> patchHttp({
    required String path,
    int retry = 0,
    String? parameter,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    String? contentType,
    Map<String, dynamic>? header,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        // await baseStorage.openBox(StorageConstants.base);
        final Map<String, dynamic> dioHeader = _dio.options.headers;
        final Map<String, dynamic> headers = {
          'content-type': ContentType.json.mimeType,
          'Accept': ContentType.json.mimeType,
        };
        if (header != null) {
          headers.addAll(header);
        }
        dioHeader.addAll(headers);
        _dio.options.headers = dioHeader;
        final Response response = await _requestPatch(
          retry: retry,
          path: path,
          content: content,
          onSendProgress: onSendProgress,
          queryParameter: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());

        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  @override
  Future<Either<NetworkError, dynamic>> delete(
      {required String deletePath,
      int retry = 0,
      Map<String, dynamic>? queryParameter,
      Map<String, dynamic>? header}) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        // await baseStorage.openBox(StorageConstants.base);
        final Map<String, dynamic> dioHeader = _dio.options.headers;
        final Map<String, dynamic> headers = {
          'content-type': ContentType.json.mimeType,
          'Accept': ContentType.json.mimeType,
        };
        if (header != null) {
          headers.addAll(header);
        }
        dioHeader.addAll(headers);
        _dio.options.headers = dioHeader;
        final Response response = await _requestDelete(
          retry: retry,
          path: deletePath,
          queryParameter: queryParameter,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());

        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  @override
  Future<Either<NetworkError, dynamic>> request({
    required String path,
    int retry = 0,
    String? parameter,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    String? contentType,
    Map<String, dynamic>? header,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
    NetworkOptions? networkOptions,
    Duration? networkReceiveTimeout,
    Duration? networkSendTimeout,
  }) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        // await baseStorage.openBox(StorageConstants.base);
        final Map<String, dynamic> dioHeader = _dio.options.headers;
        final Map<String, dynamic> headers = {
          'content-type': ContentType.json.mimeType,
          'Accept': ContentType.json.mimeType,
        };
        if (header != null) {
          headers.addAll(header);
        }
        dioHeader.addAll(headers);
        _dio.options.headers = dioHeader;
        Options? options;
        if (networkOptions != null) {
          options = Options(
            contentType: networkOptions.contentType ?? this.contentType,
            extra: networkOptions.extra ?? extra,
            followRedirects: networkOptions.followRedirects ?? followRedirects,
            headers: networkOptions.headers ?? this.headers,
            listFormat: networkOptions.listFormat ?? listFormat,
            maxRedirects: networkOptions.maxRedirects ?? maxRedirects,
            method: networkOptions.method ?? method,
            receiveDataWhenStatusError:
                networkOptions.receiveDataWhenStatusError ??
                    receiveDataWhenStatusError,
            receiveTimeout: networkReceiveTimeout ?? receiveTimeout,
            requestEncoder: networkOptions.requestEncoder ?? requestEncoder,
            responseDecoder: networkOptions.responseDecoder ?? responseDecoder,
            responseType: networkOptions.responseType ?? responseType,
            sendTimeout: networkSendTimeout ?? sendTimeout,
            validateStatus: networkOptions.validateStatus ?? validateStatus,
          );
        }
        final Response response = await _requestDynamic(
          path: path,
          retry: retry,
          content: content,
          evaluator: evaluator,
          onReceiveProgress: onReceiveProgress,
          onSendProgress: onSendProgress,
          options: options,
          queryParameter: queryParameter,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());

        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  @override
  Future<Either<NetworkError, dynamic>> download({
    required String url,
    required String downloadPath,
    required String fileName,
    Map<String, dynamic>? queryParameter,
    Map<String, dynamic>? header,
    Function(int, int)? onReceiveProgress,
  }) async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      try {
        // await baseStorage.openBox(StorageConstants.base);
        final Directory savedDir = Directory(downloadPath);
        final bool hasExisted = await savedDir.exists();

        if (!hasExisted) {
          await savedDir.create(recursive: true);
        }
        final Map<String, dynamic> dioHeader = _dio.options.headers;
        final Map<String, dynamic> headers = {
          'Accept': ContentType.binary.mimeType,
        };
        if (header != null) {
          headers.addAll(header);
        }
        dioHeader.addAll(headers);
        _dio.options.headers = dioHeader;

        final Response response = await _dio.download(
          url,
          '$downloadPath/$fileName',
          queryParameters: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
        return right(response.data);
      } on DioError catch (e) {
        debugPrint('================ERROR==================');
        debugPrint(e.type.toString());
        switch (e.type) {
          case DioErrorType.connectionTimeout:
            return left(const NetworkError.timeout());
          case DioErrorType.unknown:
            return left(NetworkError.other(e: e));
          default:
            return left(NetworkError.serverError(response: e.response));
        }
      }
    } else {
      return left(const NetworkError.noInternet());
    }
  }

  Future<Response> _requestGet({
    required int retry,
    required String path,
    Map<String, dynamic>? queryParameter,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    try {
      return await _dio.get(
        path,
        queryParameters: queryParameter,
        onReceiveProgress: onReceiveProgress,
      );
    } on DioError catch (e) {
      final bool shouldRetry = evaluator != null
          ? await evaluator(e, retry)
          : await _defaultRetryEvaluator(e, retry);
      if (shouldRetry && retry > 0) {
        await Future.delayed(const Duration(milliseconds: 300));
        return await _requestGet(
          retry: retry - 1,
          path: path,
          queryParameter: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
      }
      rethrow;
    }
  }

  Future<Response> _requestPost({
    required String path,
    required int retry,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    try {
      return await _dio.post(
        path,
        queryParameters: queryParameter,
        data: content,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
    } on DioError catch (e) {
      final bool shouldRetry = evaluator != null
          ? await evaluator(e, retry)
          : await _defaultRetryEvaluator(e, retry);
      if (shouldRetry && retry > 0) {
        await Future.delayed(const Duration(milliseconds: 300));
        return await _requestPost(
          retry: retry - 1,
          path: path,
          queryParameter: queryParameter,
          content: content,
          onSendProgress: onSendProgress,
          onReceiveProgress: onReceiveProgress,
        );
      }
      rethrow;
    }
  }

  Future<Response> _requestPut({
    required String path,
    required int retry,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    try {
      return await _dio.put(
        path,
        queryParameters: queryParameter,
        data: content,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
    } on DioError catch (e) {
      final bool shouldRetry = evaluator != null
          ? await evaluator(e, retry)
          : await _defaultRetryEvaluator(e, retry);
      if (shouldRetry && retry > 0) {
        await Future.delayed(const Duration(milliseconds: 300));
        return await _requestPut(
          retry: retry - 1,
          path: path,
          content: content,
          onSendProgress: onSendProgress,
          queryParameter: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
      }
      rethrow;
    }
  }

  Future<Response> _requestPatch({
    required String path,
    required int retry,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    RetryEvaluator? evaluator,
  }) async {
    try {
      return await _dio.patch(
        path,
        queryParameters: queryParameter,
        data: content,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
    } on DioError catch (e) {
      final bool shouldRetry = evaluator != null
          ? await evaluator(e, retry)
          : await _defaultRetryEvaluator(e, retry);
      if (shouldRetry && retry > 0) {
        await Future.delayed(const Duration(milliseconds: 300));
        return await _requestPatch(
          retry: retry - 1,
          path: path,
          content: content,
          onSendProgress: onSendProgress,
          queryParameter: queryParameter,
          onReceiveProgress: onReceiveProgress,
        );
      }
      rethrow;
    }
  }

  Future<Response> _requestDynamic({
    required String path,
    required int retry,
    Map<String, dynamic>? queryParameter,
    dynamic content,
    Function(int, int)? onSendProgress,
    Function(int, int)? onReceiveProgress,
    Options? options,
    RetryEvaluator? evaluator,
  }) async {
    try {
      return await _dio.request(
        path,
        data: content,
        queryParameters: queryParameter,
        options: options,
        onSendProgress: onSendProgress,
        onReceiveProgress: onReceiveProgress,
      );
    } on DioError catch (e) {
      final bool shouldRetry = evaluator != null
          ? await evaluator(e, retry)
          : await _defaultRetryEvaluator(e, retry);
      if (shouldRetry && retry > 0) {
        await Future.delayed(const Duration(milliseconds: 300));
        return await _requestDynamic(
          retry: retry - 1,
          path: path,
          content: content,
          queryParameter: queryParameter,
        );
      }
      rethrow;
    }
  }

  Future<Response> _requestDelete({
    required String path,
    required int retry,
    Map<String, dynamic>? queryParameter,
    Map<String, dynamic>? content,
    RetryEvaluator? evaluator,
  }) async {
    try {
      return await _dio.delete(
        path,
        queryParameters: queryParameter,
        data: content,
      );
    } on DioError catch (e) {
      final bool shouldRetry = evaluator != null
          ? await evaluator(e, retry)
          : await _defaultRetryEvaluator(e, retry);
      if (shouldRetry && retry > 0) {
        await Future.delayed(const Duration(milliseconds: 300));
        return await _requestDelete(
          retry: retry - 1,
          path: path,
          content: content,
          queryParameter: queryParameter,
        );
      }
      rethrow;
    }
  }

  FutureOr<bool> _defaultRetryEvaluator(DioError error, int attempt) {
    bool shouldRetry;
    if (error.type == DioErrorType.badResponse) {
      final statusCode = error.response?.statusCode;
      shouldRetry = statusCode != null ? isRetryable(statusCode) : true;
    } else {
      shouldRetry = error.type != DioErrorType.cancel;
    }
    return shouldRetry;
  }

  @override
  HttpClientAdapter get httpClientAdapter => _dio.httpClientAdapter;

  @override
  IOHttpClientAdapter get httpDefaultClientAdapter =>
      _dio.httpClientAdapter as IOHttpClientAdapter;

  @override
  Dio getDio() {
    return _dio;
  }

  @override
  Future<Either<NetworkError, bool>> checkInternet() async {
    final connectivityResult = await connectivity.checkConnectivity();
    if (connectivityResult != ConnectivityResult.none) {
      return right(true);
    }
    return left(const NetworkError.noInternet());
  }
}
